/* PeerCoordinator - Coordinates which peers do what (up and downloading).
   Copyright (C) 2003 Mark J. Wielaard

   This file is part of Snark.
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

package org.klomp.snark;

import java.util.*;
import java.io.IOException;

/**
 * Coordinates what peer does what.
 */
public class PeerCoordinator implements PeerListener
{
  final MetaInfo metainfo;
  final Storage storage;

  // package local for access by CheckDownLoadersTask
  final static long CHECK_PERIOD = 30*1000; // 30 seconds
  final static int MAX_CONNECTIONS = 16;
  final static int MAX_UPLOADERS = 4;
  final static int MAX_DOWNLOADERS = MAX_CONNECTIONS;
  int uploaders = 0;
  int downloaders = 0;

  private long uploaded;
  private long downloaded;

  // synchronize on this when changing peers or downloaders
  final List peers = new ArrayList();

  /** Timer to handle all periodical tasks. */
  private final Timer timer = new Timer(true);

  private final byte[] id;

  // Some random wanted pieces
  private final List wantedPieces;

  public PeerCoordinator(byte[] id, MetaInfo metainfo, Storage storage)
  {
    this.id = id;
    this.metainfo = metainfo;
    this.storage = storage;

    // Make a random list of piece numbers
    wantedPieces = new ArrayList();
    BitField bitfield = storage.getBitField();
    for(int i = 0; i < metainfo.getPieces(); i++)
      if (!bitfield.get(i))
	wantedPieces.add(new Integer(i));
    Collections.shuffle(wantedPieces);

    // Install a timer to check the uploaders.
    timer.schedule(new PeerCheckerTask(this), CHECK_PERIOD, CHECK_PERIOD);
  }

  public byte[] getID()
  {
    return id;
  }

  public boolean completed()
  {
    return storage.complete();
  }

  /**
   * Returns how many bytes are still needed to get the complete file.
   */
  public long getLeft()
  {
    // XXX - Only an approximation.
    return storage.needed() * metainfo.getPieceLength(0);
  }

  /**
   * Returns the total number of uploaded bytes of all peers.
   */
  public long getUploaded()
  {
    return uploaded;
  }

  /**
   * Returns the total number of downloaded bytes of all peers.
   */
  public long getDownloaded()
  {
    return downloaded;
  }

  public MetaInfo getMetaInfo()
  {
    return metainfo;
  }

  public boolean needPeers()
  {
    synchronized(peers)
      {
	return peers.size() < MAX_CONNECTIONS;
      }
  }

  public void connected(Peer peer)
  {
    synchronized (peers)
      {
	if (peerIDInList(peer.getPeerID(), peers))
	  {
	    if (Snark.debug)
	      Snark.debug("Already connected to: " + peer);
	    peer.disconnect(false); // Don't deregister this connection/peer.
	  }
	else
	  {
	    if (Snark.debug)
	      Snark.debug("New connection to peer: " + peer);

	    // Add it to the beginning of the list.
	    // And try to optimistically make it a up/downloader.
	    peers.add(0, peer);
	    addUpDownloader();
	  }
      }
  }

  private static boolean peerIDInList(PeerID pid, List peers)
  {
    Iterator it = peers.iterator();
    while (it.hasNext())
      if (pid.sameID(((Peer)it.next()).getPeerID()))
	return true;
    return false;
  }

  public void addPeer(final Peer peer)
  {
    synchronized(peers)
      {
	if (!peer.isConnected() && peers.size() < MAX_CONNECTIONS)
	{
	  final PeerListener listener = this;
	  Runnable r = new Runnable()
	    {
	      public void run()
	      {
		// Run the peer with us as listener and the current bitfield.
		peer.runConnection(listener, storage.getBitField());
	      }
	    };
	  new Thread(r, peer.toString()).start();
	}
	else
	  if (Snark.debug)
	    if (peer.isConnected())
	      Snark.debug("Add peer already connected: " + peer);
	    else
	      Snark.debug("MAX_CONNECTIONS = " + MAX_CONNECTIONS
			  + " not accepting extra peer: " + peer);
      }
  }

  // Optimistically unchoke. Should be called with peers synchronized
  void addUpDownloader()
  {
    List removed = new ArrayList();
    Iterator it = peers.iterator();
    while (it.hasNext())
      {
	Peer peer = (Peer)it.next();
	boolean remove = false;
	if (uploaders < MAX_UPLOADERS
	    && !peer.isUpLoading()
	    && peer.isInterested())
	  {
	    if (Snark.debug)
	      Snark.debug("Add new uploader: " + peer);
	    peer.upload(true);
	    uploaders++;
	    remove = true;
	  }
	
	if (downloaders < MAX_DOWNLOADERS
	    && !peer.isDownLoading()
	    && peer.isInteresting()
	    && !peer.isChoked())
	  {
	    if (Snark.debug)
	      Snark.debug("Add new downloader: " + peer);
	    peer.download(true);
	    downloaders++;
	    remove = true;
	  }

	// Remove peer now, so we can add it back at the end later.
	if (remove)
	  {
	    removed.add(peer);
	    it.remove();
	  }
      }
    // Put peers back at the end of the list.
    peers.addAll(removed);
  }

  public byte[] getBitMap()
  {
    return storage.getBitField().getFieldBytes();
  }

  /**
   * Returns true if we don't have the given piece yet.
   */
  public boolean gotHave(Peer peer, int piece)
  {
    synchronized(wantedPieces)
      {
	return wantedPieces.contains(new Integer(piece));
      }
  }

  /**
   * Returns true if the given bitfield contains at least one piece we
   * are interested in.
   */
  public boolean gotBitField(Peer peer, BitField bitfield)
  {
    synchronized(wantedPieces)
      {
	Iterator it = wantedPieces.iterator();
	while (it.hasNext())
	  {
	    int i = ((Integer)it.next()).intValue();
	    if (bitfield.get(i))
	      return true;
	  }
      }
    return false;
  }

  /**
   * Returns one of pieces in the given BitField that is still wanted or
   * -1 if none of the given pieces are wanted.
   */
  public int wantPiece(Peer peer, BitField havePieces)
  {
    synchronized(wantedPieces)
      {
	Integer piece = null;
	Iterator it = wantedPieces.iterator();
	while (piece == null && it.hasNext())
	  {
	    Integer i = (Integer)it.next();
	    if (havePieces.get(i.intValue()))
	      {
		it.remove();
		piece = i;
	      }
	  }

	if (piece == null)
	  return -1;

	// We add it back at the back of the list. It will be removed
	// if gotPiece is called later. This means that the last
	// couple of pieces might very well be asked from multiple
	// peers but that is OK.
	wantedPieces.add(piece);

	return piece.intValue();
      }
  }

  /**
   * Returns a byte array containing the requested piece or null of
   * the piece is unknown.
   */
  public byte[] gotRequest(Peer peer, int piece)
  {
    try
      {
	return storage.getPiece(piece);
      }
    catch (IOException ioe)
      {
	Snark.fatal("Error reading storage", ioe);
	return null; // Never reached.
      }
  }

  /**
   * Called when a peer has uploaded some bytes of a piece.
   */
  public void uploaded(Peer peer, int size)
  {
    uploaded += size;
  }

  /**
   * Returns false if the piece is no good (according to the hash).
   * In that case the peer that supplied the piece should probably be
   * blacklisted.
   */
  public boolean gotPiece(Peer peer, int piece, byte[] bs)
  {
    synchronized (wantedPieces)
      {
	try
	  {
	    if (storage.putPiece(piece, bs))
	      {
		if (!wantedPieces.remove(new Integer(piece)))
		  {
		    System.err.println("What! " + piece + " already removed?");
		    return true; // No need to announce have piece to peers.
		  }
		if (Snark.debug)
		  Snark.debug("Recv p" + piece + " " + peer);
		downloaded += metainfo.getPieceLength(piece);
	      }
	    else
	      {
		System.err.println("Got BAD piece " + piece + " from " + peer);
		return false; // No need to announce have piece to peers.
	      }
	  }
	catch (IOException ioe)
	  {
	    Snark.fatal("Error writing storage", ioe);
	  }
      }

    // Announce to the world we have it!
    synchronized(peers)
      {
	Iterator it = peers.iterator();
	while (it.hasNext())
	  {
	    Peer p = (Peer)it.next();
	    if (p.isConnected())
	      p.have(piece);
	  }
      }
    
    return true;
  }

  public void gotChoke(Peer peer, boolean choke)
  {
    if (!choke)
      {
	synchronized(peers)
	  {
	    if (downloaders < MAX_DOWNLOADERS && peer.isInteresting())
	      {
		if(!peer.isDownLoading())
		  {
		    downloaders++;
		    if (Snark.debug)
		      Snark.debug("New downloader: " + peer);
		  }
		// Make sure we actually (re)download.
		peer.download(true);
	      }
	  }
      }
  }

  public void gotInterest(Peer peer, boolean interest)
  {
    if (interest)
      {
	synchronized(peers)
	  {
	    if (uploaders < MAX_UPLOADERS)
	      {
		if(!peer.isUpLoading())
		  {
		    uploaders++;
		    if (Snark.debug)
		      Snark.debug("New uploader: " + peer);
		  }
		// Make sure we actually (re)upload.
		peer.upload(true);
	      }
	  }
      }
  }

  public void disconnected(Peer peer)
  {
    synchronized(peers)
      {
	// Make sure it is no longer in our lists
	if (peers.remove(peer))
	  {
	    if (Snark.debug)
	      Snark.debug("Disconnected " + peer);

	    if (peer.isUpLoading())
	      uploaders--;
	    if (peer.isDownLoading())
	      downloaders--;
	    
	    // Unchoke some random other peer
	    addUpDownloader();
	  }
      }
  }
}
